import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import yaml
from pydantic import BaseModel, ValidationError


class CostRange(BaseModel):
    range: List[Optional[int]]
    cost: float


class ModelConfig(BaseModel):
    mode: str
    max_tokens: int
    input_token_cost: Union[float, List[CostRange]]
    output_token_cost: Union[float, List[CostRange]]


class ProviderConfig(BaseModel):
    id: str
    name: str
    chat: bool
    embed: bool
    keys: Optional[List[str]] = None
    models: Optional[Dict[str, ModelConfig]] = None
    parameters: Optional[Dict[str, Any]] = None


class EngineConfig(BaseModel):
    providers: Dict[str, ProviderConfig]


def _load_providers_config() -> EngineConfig:
    # TODO read from github
    default_config_path = Path(os.path.join(os.path.dirname(__file__), "config.yaml"))
    local_config_path = Path(os.getcwd(), "config.yaml")

    def _merge_configs(config1, config2):
        for key in config2:
            if key in config1:
                if isinstance(config1[key], dict) and isinstance(config2[key], dict):
                    _merge_configs(config1[key], config2[key])
                elif isinstance(config1[key], list) and isinstance(config2[key], list):
                    config1[key].extend(config2[key])
                else:
                    config1[key] = config2[key]
            else:
                config1[key] = config2[key]
        return config1

    try:
        default_config_data = yaml.safe_load(default_config_path.read_text())
        local_config_data = (
            yaml.safe_load(local_config_path.read_text())
            if local_config_path.exists()
            else {}
        )
        config_data = _merge_configs(default_config_data, local_config_data)
        return EngineConfig(**config_data)
    except yaml.YAMLError as e:
        raise RuntimeError(f"Error parsing YAML configuration: {e}")
    except ValidationError as e:
        raise RuntimeError(f"Error in configuration data: {e}")



import asyncio
import json
import os
import time
import uuid
from typing import (
    Any,
    AsyncGenerator,
    Coroutine,
    Dict,
    Generator,
    List,
    Optional,
    Union,
)

import requests
from llmstudio_core.exceptions import ProviderError
from llmstudio_core.providers.provider import ChatRequest, ProviderCore, provider
from openai.types.chat import ChatCompletionChunk
from openai.types.chat.chat_completion_chunk import (
    Choice,
    ChoiceDelta,
    ChoiceDeltaToolCall,
    ChoiceDeltaToolCallFunction,
)
from pydantic import BaseModel, ValidationError


class OpenAIToolParameter(BaseModel):
    type: str
    description: Optional[str] = None


class OpenAIToolParameters(BaseModel):
    type: str
    properties: Dict[str, OpenAIToolParameter]
    required: List[str]


class OpenAIToolFunction(BaseModel):
    name: str
    description: str
    parameters: OpenAIToolParameters


class OpenAITool(BaseModel):
    type: str
    function: OpenAIToolFunction


class VertexAIToolParameter(BaseModel):
    type: str
    description: str


class VertexAIToolParameters(BaseModel):
    type: str
    properties: Dict[str, VertexAIToolParameter]
    required: List[str]


class VertexAIFunctionDeclaration(BaseModel):
    name: str
    description: str
    parameters: VertexAIToolParameters


class VertexAI(BaseModel):
    function_declarations: List[VertexAIFunctionDeclaration]


@provider
class VertexAIProvider(ProviderCore):
    def __init__(self, config, **kwargs):
        super().__init__(config, **kwargs)
        self.API_KEY = self.API_KEY if self.API_KEY else os.getenv("GOOGLE_API_KEY")

    @staticmethod
    def _provider_config_name():
        return "vertexai"

    def validate_request(self, request: ChatRequest):
        return ChatRequest(**request)

    async def agenerate_client(
        self, request: ChatRequest
    ) -> Coroutine[Any, Any, Generator]:
        """Initialize Vertex AI"""

        try:
            # Init genai
            url = f"https://generativelanguage.googleapis.com/v1beta/models/{request.model}:streamGenerateContent?alt=sse"
            headers = {
                "Content-Type": "application/json",
                "x-goog-api-key": self.API_KEY,
            }

            # Convert the chat input into VertexAI format
            tool_payload = self.process_tools(request.parameters.get("tools"))
            message = self.convert_input_to_vertexai(request.chat_input, tool_payload)

            # Generate content
            return await asyncio.to_thread(
                requests.post, url, headers=headers, json=message, stream=True
            )

        except Exception as e:
            raise ProviderError(str(e))

    def generate_client(self, request: ChatRequest) -> Coroutine[Any, Any, Generator]:
        """Initialize Vertex AI"""

        try:
            # Init genai
            url = f"https://generativelanguage.googleapis.com/v1beta/models/{request.model}:streamGenerateContent?alt=sse"
            headers = {
                "Content-Type": "application/json",
                "x-goog-api-key": self.API_KEY,
            }

            # Convert the chat input into VertexAI format
            tool_payload = self.process_tools(request.parameters.get("tools"))
            message = self.convert_input_to_vertexai(request.chat_input, tool_payload)

            # Generate content
            return requests.post(url, headers=headers, json=message, stream=True)

        except Exception as e:
            raise ProviderError(str(e))

    def parse_response(self, response: AsyncGenerator[Any, None], **kwargs) -> Any:

        for chunk in response.iter_content(chunk_size=None):

            chunk = json.loads(chunk.decode("utf-8").lstrip("data: "))
            chunk = chunk.get("candidates")[0].get("content")

            if not chunk:
                continue

            # Check if it is a function call
            if (
                "functionCall" in chunk["parts"][0]
                and chunk["parts"][0]["functionCall"] is not None
            ):
                first_chunk = ChatCompletionChunk(
                    id="chatcmpl-9woLM1b1qGErhTbXA3UBQf2FhUAho",
                    choices=[
                        Choice(
                            delta=ChoiceDelta(
                                content=None,
                                function_call=None,
                                role="assistant",
                                tool_calls=None,
                            ),
                            index=0,
                        )
                    ],
                    created=int(time.time()),
                    model=kwargs.get("request").model,
                    object="chat.completion.chunk",
                    usage=None,
                )
                yield first_chunk.model_dump()

                for index, functioncall in enumerate(chunk["parts"]):

                    name_chunk = ChatCompletionChunk(
                        id=str(uuid.uuid4()),
                        choices=[
                            Choice(
                                delta=ChoiceDelta(
                                    role="assistant",
                                    tool_calls=[
                                        ChoiceDeltaToolCall(
                                            index=index,
                                            id="call_" + str(uuid.uuid4())[:29],
                                            function=ChoiceDeltaToolCallFunction(
                                                name=functioncall["functionCall"][
                                                    "name"
                                                ],
                                                arguments="",
                                                type="function",
                                            ),
                                        )
                                    ],
                                ),
                                finish_reason=None,
                                index=index,
                            )
                        ],
                        created=int(time.time()),
                        model=kwargs.get("request").model,
                        object="chat.completion.chunk",
                    )
                    yield name_chunk.model_dump()

                    args_chunk = ChatCompletionChunk(
                        id=str(uuid.uuid4()),
                        choices=[
                            Choice(
                                delta=ChoiceDelta(
                                    tool_calls=[
                                        ChoiceDeltaToolCall(
                                            index=index,
                                            function=ChoiceDeltaToolCallFunction(
                                                arguments=json.dumps(
                                                    functioncall["functionCall"]["args"]
                                                ),
                                            ),
                                        )
                                    ],
                                ),
                                finish_reason=None,
                                index=index,
                            )
                        ],
                        created=int(time.time()),
                        model=kwargs.get("request").model,
                        object="chat.completion.chunk",
                    )
                    yield args_chunk.model_dump()

                final_chunk = ChatCompletionChunk(
                    id=str(uuid.uuid4()),
                    choices=[
                        Choice(
                            delta=ChoiceDelta(),
                            finish_reason="tool_calls",
                            index=0,
                        )
                    ],
                    created=int(time.time()),
                    model=kwargs.get("request").model,
                    object="chat.completion.chunk",
                )
                yield final_chunk.model_dump()
            # Check if it is a normal call
            elif chunk.get("parts")[0].get("text"):
                # Parse google chunk response into ChatCompletionChunk
                yield ChatCompletionChunk(
                    id=str(uuid.uuid4()),
                    choices=[
                        Choice(
                            delta=ChoiceDelta(
                                content=chunk.get("parts")[0].get("text"),
                                role="assistant",
                            ),
                            finish_reason=None,
                            index=0,
                        )
                    ],
                    created=int(time.time()),
                    model=kwargs.get("request").model,
                    object="chat.completion.chunk",
                ).model_dump()

                # Create the closing chunk
                yield ChatCompletionChunk(
                    id=str(uuid.uuid4()),
                    choices=[
                        Choice(delta=ChoiceDelta(), finish_reason="stop", index=0)
                    ],
                    created=int(time.time()),
                    model=kwargs.get("request").model,
                    object="chat.completion.chunk",
                ).model_dump()

    async def aparse_response(
        self, response: AsyncGenerator, **kwargs
    ) -> AsyncGenerator[str, None]:
        for chunk in response.iter_content(chunk_size=None):

            chunk = json.loads(chunk.decode("utf-8").lstrip("data: "))
            chunk = chunk.get("candidates")[0].get("content")

            # Check if it is a function call
            if (
                "functionCall" in chunk["parts"][0]
                and chunk["parts"][0]["functionCall"] is not None
            ):
                first_chunk = ChatCompletionChunk(
                    id="chatcmpl-9woLM1b1qGErhTbXA3UBQf2FhUAho",
                    choices=[
                        Choice(
                            delta=ChoiceDelta(
                                content=None,
                                function_call=None,
                                role="assistant",
                                tool_calls=None,
                            ),
                            index=0,
                        )
                    ],
                    created=int(time.time()),
                    model=kwargs.get("request").model,
                    object="chat.completion.chunk",
                    usage=None,
                )
                yield first_chunk.model_dump()

                for index, functioncall in enumerate(chunk["parts"]):

                    name_chunk = ChatCompletionChunk(
                        id=str(uuid.uuid4()),
                        choices=[
                            Choice(
                                delta=ChoiceDelta(
                                    role="assistant",
                                    tool_calls=[
                                        ChoiceDeltaToolCall(
                                            index=index,
                                            id="call_" + str(uuid.uuid4())[:29],
                                            function=ChoiceDeltaToolCallFunction(
                                                name=functioncall["functionCall"][
                                                    "name"
                                                ],
                                                arguments="",
                                                type="function",
                                            ),
                                        )
                                    ],
                                ),
                                finish_reason=None,
                                index=index,
                            )
                        ],
                        created=int(time.time()),
                        model=kwargs.get("request").model,
                        object="chat.completion.chunk",
                    )
                    yield name_chunk.model_dump()

                    args_chunk = ChatCompletionChunk(
                        id=str(uuid.uuid4()),
                        choices=[
                            Choice(
                                delta=ChoiceDelta(
                                    tool_calls=[
                                        ChoiceDeltaToolCall(
                                            index=index,
                                            function=ChoiceDeltaToolCallFunction(
                                                arguments=json.dumps(
                                                    functioncall["functionCall"]["args"]
                                                ),
                                            ),
                                        )
                                    ],
                                ),
                                finish_reason=None,
                                index=index,
                            )
                        ],
                        created=int(time.time()),
                        model=kwargs.get("request").model,
                        object="chat.completion.chunk",
                    )
                    yield args_chunk.model_dump()

                final_chunk = ChatCompletionChunk(
                    id=str(uuid.uuid4()),
                    choices=[
                        Choice(
                            delta=ChoiceDelta(),
                            finish_reason="tool_calls",
                            index=0,
                        )
                    ],
                    created=int(time.time()),
                    model=kwargs.get("request").model,
                    object="chat.completion.chunk",
                )
                yield final_chunk.model_dump()
            # Check if it is a normal call
            elif chunk.get("parts")[0].get("text"):
                # Parse google chunk response into ChatCompletionChunk
                yield ChatCompletionChunk(
                    id=str(uuid.uuid4()),
                    choices=[
                        Choice(
                            delta=ChoiceDelta(
                                content=chunk.get("parts")[0].get("text"),
                                role="assistant",
                            ),
                            finish_reason=None,
                            index=0,
                        )
                    ],
                    created=int(time.time()),
                    model=kwargs.get("request").model,
                    object="chat.completion.chunk",
                ).model_dump()

                # Create the closing chunk
                yield ChatCompletionChunk(
                    id=str(uuid.uuid4()),
                    choices=[
                        Choice(delta=ChoiceDelta(), finish_reason="stop", index=0)
                    ],
                    created=int(time.time()),
                    model=kwargs.get("request").model,
                    object="chat.completion.chunk",
                ).model_dump()

    def convert_input_to_vertexai(
        self, input_data: Union[Dict, str, List[Dict]], tool_payload: Optional[Any]
    ) -> Dict:
        """
        Converts OpenAI formatted input to VertexAI format.

        Args:
            input_data (Union[Dict, str, List[Dict]]): The input data in OpenAI format.
            tool_payload (Optional[Any]): The tool payload for the request.

        Returns:
            Dict: The converted input in VertexAI format.
        """
        if isinstance(input_data, dict) and "input" in input_data:
            return self._handle_simple_string_input(input_data["input"], tool_payload)

        if isinstance(input_data, str):
            return self._handle_simple_string_input(input_data, tool_payload)

        if isinstance(input_data, list):
            return self._convert_list_input_to_vertexai(input_data, tool_payload)

        raise ValueError("Invalid input type. Expected dict, str, or list.")

    def _handle_simple_string_input(
        self, input_data: str, tool_payload: Optional[Any]
    ) -> Dict:
        """
        Handles simple string input and converts it to VertexAI format.

        Args:
            input_data (str): The input data as a simple string.
            tool_payload (Optional[Any]): The tool payload for the request.

        Returns:
            Dict: The converted input in VertexAI format.
        """
        return self._initialize_vertexai_message(
            user_message=input_data, tool_payload=tool_payload
        )

    def _convert_list_input_to_vertexai(
        self, input_data: List[Dict], tool_payload: Optional[Any]
    ) -> Dict:
        """
        Converts a list of messages in OpenAI format to VertexAI format.

        Args:
            input_data (List[Dict]): The input data as a list of messages.
            tool_payload (Optional[Any]): The tool payload for the request.

        Returns:
            Dict: The converted input in VertexAI format.
        """
        vertexai_format = self._initialize_vertexai_message(tool_payload=tool_payload)
        for message in input_data:
            role = message.get("role")
            if role == "system":
                vertexai_format["system_instruction"]["parts"]["text"] = (
                    message["content"] or "You are a helpful assistant"
                )
            elif role == "user":
                vertexai_format["contents"].append(
                    {"role": "user", "parts": [{"text": message["content"]}]}
                )
            elif role == "assistant":
                if message["content"] is None and "tool_calls" in message:
                    tool_call = message["tool_calls"][0]
                    vertexai_format["contents"].append(
                        {
                            "role": "model",
                            "parts": [
                                {
                                    "functionCall": {
                                        "name": tool_call["function"]["name"],
                                        "args": json.loads(
                                            tool_call["function"]["arguments"]
                                        ),
                                    }
                                }
                            ],
                        }
                    )
                else:
                    vertexai_format["contents"].append(
                        {"role": "model", "parts": [{"text": message["content"]}]}
                    )
            elif role == "tool":
                function_name = message["name"]
                response = message["content"]
                vertexai_format["system_instruction"]["parts"][
                    "text"
                ] += f"\nYou have called {function_name} and got the following response: {response}."
            else:
                raise ValueError(
                    f"Invalid role: {role}. Expected 'system', 'user', 'assistant', or 'tool'."
                )
        return vertexai_format

    def _initialize_vertexai_message(
        self, user_message: Optional[str] = None, tool_payload: Optional[Any] = None
    ) -> Dict:
        """
        Initializes the basic structure of a VertexAI message.

        Args:
            user_message (Optional[str]): The user's message to include in the request.
            tool_payload (Optional[Any]): The tool payload for the request.

        Returns:
            Dict: The initialized VertexAI message structure.
        """
        message_format = {
            "system_instruction": {"parts": {"text": "You are a helpful assistant"}},
            "contents": [],
            "tools": tool_payload,
            "tool_config": {"function_calling_config": {"mode": "AUTO"}},
        }
        if user_message:
            message_format["contents"].append(
                {"role": "user", "parts": [{"text": user_message}]}
            )
        return message_format

    def process_tools(
        self, tools: Optional[Union[List[Dict], Dict]]
    ) -> Optional[VertexAI]:
        if tools is None:
            return None

        try:
            # Try to parse as OpenAI format
            parsed_tools = (
                [OpenAITool(**tool) for tool in tools]
                if isinstance(tools, list)
                else [OpenAITool(**tools)]
            )
            # Convert to VertexAI format
            function_declarations = []
            for tool in parsed_tools:
                function = tool.function
                properties = {
                    name: VertexAIToolParameter(
                        type=param.type, description=param.description or ""
                    )
                    for name, param in function.parameters.properties.items()
                }
                function_decl = VertexAIFunctionDeclaration(
                    name=function.name,
                    description=function.description,
                    parameters=VertexAIToolParameters(
                        type=function.parameters.type,
                        properties=properties,
                        required=function.parameters.required,
                    ),
                )
                function_declarations.append(function_decl)
            return VertexAI(function_declarations=function_declarations).model_dump()
        except ValidationError:
            # If the format is not OpenAI, attempt to validate as VertexAI format
            try:
                return VertexAI(**tools).model_dump()
            except ValidationError:
                # If it fails to validate as VertexAI, throw an error
                raise ValueError(
                    "Invalid tool format. Tool data must be in OpenAI or VertexAI format."
                )



import asyncio
import os
from typing import Any, AsyncGenerator, Coroutine, Generator

import openai
from llmstudio_core.exceptions import ProviderError
from llmstudio_core.providers.provider import ChatRequest, ProviderCore, provider
from openai import OpenAI


@provider
class OpenAIProvider(ProviderCore):
    def __init__(self, config, **kwargs):
        super().__init__(config, **kwargs)
        self.API_KEY = self.API_KEY if self.API_KEY else os.getenv("OPENAI_API_KEY")
        self._client = OpenAI(api_key=self.API_KEY)

    @staticmethod
    def _provider_config_name():
        return "openai"

    def validate_request(self, request: ChatRequest):
        return ChatRequest(**request)

    async def agenerate_client(
        self, request: ChatRequest
    ) -> Coroutine[Any, Any, Generator]:
        """Generate an OpenAI client"""

        try:
            return await asyncio.to_thread(
                self._client.chat.completions.create,
                model=request.model,
                messages=(
                    [{"role": "user", "content": request.chat_input}]
                    if isinstance(request.chat_input, str)
                    else request.chat_input
                ),
                stream=True,
                **request.parameters,
            )
        except openai._exceptions.APIError as e:
            raise ProviderError(str(e))

    def generate_client(self, request: ChatRequest) -> Generator:
        """Generate an OpenAI client"""

        try:
            return self._client.chat.completions.create(
                model=request.model,
                messages=(
                    [{"role": "user", "content": request.chat_input}]
                    if isinstance(request.chat_input, str)
                    else request.chat_input
                ),
                stream=True,
                **request.parameters,
            )
        except openai._exceptions.APIError as e:
            raise ProviderError(str(e))

    async def aparse_response(
        self, response: AsyncGenerator, **kwargs
    ) -> AsyncGenerator[str, None]:
        for chunk in response:
            yield chunk.model_dump()

    def parse_response(self, response: Generator, **kwargs) -> Generator:
        for chunk in response:
            yield chunk.model_dump()



import ast
import asyncio
import json
import os
import time
import uuid
from typing import Any, AsyncGenerator, Generator, Union

import openai
from llmstudio_core.exceptions import ProviderError
from llmstudio_core.providers.provider import ChatRequest, ProviderCore, provider
from openai import AzureOpenAI, OpenAI
from openai.types.chat import ChatCompletionChunk
from openai.types.chat.chat_completion_chunk import (
    Choice,
    ChoiceDelta,
    ChoiceDeltaFunctionCall,
    ChoiceDeltaToolCall,
    ChoiceDeltaToolCallFunction,
)


@provider
class AzureProvider(ProviderCore):
    def __init__(
        self, config, api_key=None, api_endpoint=None, api_version=None, base_url=None
    ):
        super().__init__(config)
        self.API_KEY = api_key or os.getenv("AZURE_API_KEY")
        self.API_ENDPOINT = api_endpoint
        self.API_VERSION = api_version or os.getenv("AZURE_API_VERSION")
        self.BASE_URL = base_url
        self.is_llama = False
        self.has_tools_functions = False

        if self.BASE_URL and (self.API_ENDPOINT is None):
            self._client = OpenAI(
                api_key=self.API_KEY,
                base_url=self.BASE_URL,
            )
        else:
            self._client = AzureOpenAI(
                api_key=self.API_KEY,
                azure_endpoint=self.API_ENDPOINT,
                api_version=self.API_VERSION,
            )

    @staticmethod
    def _provider_config_name():
        return "azure"

    def validate_request(self, request: ChatRequest):
        return ChatRequest(**request)

    async def agenerate_client(self, request: ChatRequest) -> Any:
        """Generate an AzureOpenAI client"""

        self.is_llama = "llama" in request.model.lower()
        self.is_openai = "gpt" in request.model.lower()
        self.has_tools = request.parameters.get("tools") is not None
        self.has_functions = request.parameters.get("functions") is not None

        try:
            messages = self.prepare_messages(request)

            # Prepare the optional tool-related arguments
            tool_args = {}
            if not self.is_llama and self.has_tools and self.is_openai:
                tool_args = {
                    "tools": request.parameters.get("tools"),
                    "tool_choice": "auto" if request.parameters.get("tools") else None,
                }

            # Prepare the optional function-related arguments
            function_args = {}
            if not self.is_llama and self.has_functions and self.is_openai:
                function_args = {
                    "functions": request.parameters.get("functions"),
                    "function_call": "auto"
                    if request.parameters.get("functions")
                    else None,
                }

            # Prepare the base arguments
            base_args = {
                "model": request.model,
                "messages": messages,
                "stream": True,
            }

            # Combine all arguments
            combined_args = {
                **base_args,
                **tool_args,
                **function_args,
                **request.parameters,
            }
            # Perform the asynchronous call
            return await asyncio.to_thread(
                self._client.chat.completions.create, **combined_args
            )

        except openai._exceptions.APIConnectionError as e:
            raise ProviderError(f"There was an error reaching the endpoint: {e}")

        except openai._exceptions.APIStatusError as e:
            raise ProviderError(e.response.json())

    def generate_client(self, request: ChatRequest) -> Any:
        """Generate an AzureOpenAI client"""

        self.is_llama = "llama" in request.model.lower()
        self.is_openai = "gpt" in request.model.lower()
        self.has_tools = request.parameters.get("tools") is not None
        self.has_functions = request.parameters.get("functions") is not None

        try:
            messages = self.prepare_messages(request)

            # Prepare the optional tool-related arguments
            tool_args = {}
            if not self.is_llama and self.has_tools and self.is_openai:
                tool_args = {
                    "tools": request.parameters.get("tools"),
                    "tool_choice": "auto" if request.parameters.get("tools") else None,
                }

            # Prepare the optional function-related arguments
            function_args = {}
            if not self.is_llama and self.has_functions and self.is_openai:
                function_args = {
                    "functions": request.parameters.get("functions"),
                    "function_call": "auto"
                    if request.parameters.get("functions")
                    else None,
                }

            # Prepare the base arguments
            base_args = {
                "model": request.model,
                "messages": messages,
                "stream": True,
            }

            # Combine all arguments
            combined_args = {
                **base_args,
                **tool_args,
                **function_args,
                **request.parameters,
            }
            return self._client.chat.completions.create(**combined_args)

        except openai._exceptions.APIConnectionError as e:
            raise ProviderError(f"There was an error reaching the endpoint: {e}")

        except openai._exceptions.APIStatusError as e:
            raise ProviderError(e.response.json())

    def prepare_messages(self, request: ChatRequest):
        if self.is_llama and (self.has_tools or self.has_functions):
            user_message = self.convert_to_openai_format(request.chat_input)
            content = "<|begin_of_text|>"
            content = self.add_system_message(
                user_message,
                content,
                request.parameters.get("tools"),
                request.parameters.get("functions"),
            )
            content = self.add_conversation(user_message, content)
            return [{"role": "user", "content": content}]
        else:
            return (
                [{"role": "user", "content": request.chat_input}]
                if isinstance(request.chat_input, str)
                else request.chat_input
            )

    async def aparse_response(
        self, response: AsyncGenerator, **kwargs
    ) -> AsyncGenerator[str, None]:
        if self.is_llama and (self.has_tools or self.has_functions):
            for chunk in self.handle_tool_response(response, **kwargs):
                if chunk:
                    yield chunk
        else:
            for chunk in response:
                c = chunk.model_dump()
                if c.get("choices"):
                    yield c

    def parse_response(self, response: Generator, **kwargs) -> Any:
        if self.is_llama and (self.has_tools or self.has_functions):
            for chunk in self.handle_tool_response(response, **kwargs):
                if chunk:
                    yield chunk
        else:
            for chunk in response:
                c = chunk.model_dump()
                if c.get("choices"):
                    yield c

    def handle_tool_response(self, response: AsyncGenerator, **kwargs) -> Generator:
        """
        Asynchronously handles tool responses by parsing the content for function calls or tool activations.
        It processes the response chunks to extract and execute embedded function or tool calls, then yields
        the processed chunks or the results of these calls.

        Args:
            response: An asynchronous generator yielding response chunks from the tool or function.
            **kwargs: Additional keyword arguments that may be required for processing.

        Yields:
            An asynchronous generator of strings or processed chunks, depending on whether the response
            contains function or tool calls that need to be executed or just plain content to be returned.
        """

        function_call_buffer = ""
        saving = False
        normal_call_chunks = []
        for chunk in response:
            if chunk.choices[0].delta.content is not None:
                if (
                    "§" in chunk.choices[0].delta.content
                    or "<|python_tag|>" in chunk.choices[0].delta.content
                ):
                    saving = True

            if saving:
                function_call_buffer += chunk.choices[0].delta.content
                if chunk.choices[0].finish_reason == "stop":
                    cleaned_buffer = function_call_buffer.replace("§", "").replace(
                        "<|python_tag|>", ""
                    )
                    result_dict = eval(cleaned_buffer)

                    if self.has_functions:

                        # Create first chunk
                        first_chunk = self.create_tool_first_chunk(kwargs)
                        yield first_chunk

                        name_chunk = self.create_function_name_chunk(
                            result_dict["name"], kwargs
                        )
                        yield name_chunk

                        parameters = json.dumps(result_dict["parameters"])
                        args_chunk = self.create_function_argument_chunk(
                            parameters, kwargs
                        )
                        yield args_chunk

                        finish_chunk = self.create_function_finish_chunk(kwargs)
                        yield finish_chunk

                    if self.has_tools:
                        name_chunk = self.create_tool_name_chunk(
                            result_dict["name"], kwargs
                        )
                        yield name_chunk

                        parameters = json.dumps(result_dict["parameters"])
                        args_chunk = self.create_tool_argument_chunk(parameters, kwargs)
                        yield args_chunk

                        finish_chunk = self.create_tool_finish_chunk(kwargs)
                        yield finish_chunk

            else:
                normal_call_chunks.append(chunk)
                if chunk.choices[0].finish_reason == "stop":
                    for chunk in normal_call_chunks:
                        normal_call_chunks.append(chunk)
                if chunk.choices[0].finish_reason == "stop":
                    for chunk in normal_call_chunks:
                        yield chunk.model_dump()

    def create_tool_name_chunk(self, function_name: str, kwargs: dict) -> dict:
        return ChatCompletionChunk(
            id=str(uuid.uuid4()),
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        role="assistant",
                        tool_calls=[
                            ChoiceDeltaToolCall(
                                index=0,
                                id=str(uuid.uuid4()),
                                function=ChoiceDeltaToolCallFunction(
                                    name=function_name,
                                    arguments="",
                                    type="function",
                                ),
                            )
                        ],
                    ),
                    finish_reason=None,
                    index=0,
                )
            ],
            created=int(time.time()),
            model=kwargs.get("request").model,
            object="chat.completion.chunk",
        ).model_dump()

    def create_function_name_chunk(self, function_name: str, kwargs: dict) -> dict:
        return ChatCompletionChunk(
            id=str(uuid.uuid4()),
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=None,
                        function_call=ChoiceDeltaFunctionCall(
                            arguments="", name=function_name
                        ),
                        role="assistant",
                        tool_calls=None,
                        refusal=None,
                    ),
                    finish_reason=None,
                    index=0,
                    logprobs=None,
                )
            ],
            created=int(time.time()),
            model=kwargs.get("request").model,
            object="chat.completion.chunk",
            system_fingerprint=None,
            usage=None,
        ).model_dump()

    def create_tool_finish_chunk(self, kwargs: dict) -> dict:
        return ChatCompletionChunk(
            id=str(uuid.uuid4()),
            choices=[
                Choice(
                    delta=ChoiceDelta(),
                    finish_reason="tool_calls",
                    index=0,
                )
            ],
            created=int(time.time()),
            model=kwargs.get("request").model,
            object="chat.completion.chunk",
        ).model_dump()

    def create_tool_argument_chunk(self, content: str, kwargs: dict) -> dict:
        return ChatCompletionChunk(
            id=str(uuid.uuid4()),
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        tool_calls=[
                            ChoiceDeltaToolCall(
                                index=0,
                                function=ChoiceDeltaToolCallFunction(
                                    arguments=content,
                                ),
                            )
                        ],
                    ),
                    finish_reason=None,
                    index=0,
                )
            ],
            created=int(time.time()),
            model=kwargs.get("request").model,
            object="chat.completion.chunk",
        ).model_dump()

    def create_function_argument_chunk(self, content: str, kwargs: dict) -> dict:
        return ChatCompletionChunk(
            id=str(uuid.uuid4()),
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=None,
                        function_call=ChoiceDeltaFunctionCall(
                            arguments=content, name=None
                        ),
                        role=None,
                        tool_calls=None,
                    ),
                    finish_reason=None,
                    index=0,
                    logprobs=None,
                )
            ],
            created=int(time.time()),
            model=kwargs.get("request").model,
            object="chat.completion.chunk",
            system_fingerprint=None,
            usage=None,
        ).model_dump()

    def create_tool_first_chunk(self, kwargs: dict) -> dict:
        return ChatCompletionChunk(
            id=str(uuid.uuid4()),
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=None,
                        function_call=None,
                        role="assistant",
                        tool_calls=None,
                    ),
                    index=0,
                )
            ],
            created=int(time.time()),
            model=kwargs.get("request").model,
            object="chat.completion.chunk",
            usage=None,
        ).model_dump()

    def create_function_finish_chunk(self, kwargs: dict) -> dict:
        return ChatCompletionChunk(
            id=str(uuid.uuid4()),
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=None, function_call=None, role=None, tool_calls=None
                    ),
                    finish_reason="function_call",
                    index=0,
                    logprobs=None,
                )
            ],
            created=int(time.time()),
            model=kwargs.get("request").model,
            object="chat.completion.chunk",
            system_fingerprint=None,
            usage=None,
        ).model_dump()

    def convert_to_openai_format(self, message: Union[str, list]) -> list:
        if isinstance(message, str):
            return [{"role": "user", "content": message}]
        return message

    def add_system_message(
        self, openai_message: list, llama_message: str, tools: list, functions: list
    ) -> str:
        system_message = ""
        system_message_found = False
        for message in openai_message:
            if message["role"] == "system" and message["content"] is not None:
                system_message_found = True
                system_message = f"""
        <|start_header_id|>system<|end_header_id|>
        {message['content']}
        """
        if not system_message_found:
            system_message = """
      <|start_header_id|>system<|end_header_id|>
      You are a helpful AI assistant.
      """

        if tools:
            system_message = system_message + self.add_tool_instructions(tools)

        if functions:
            system_message = system_message + self.add_function_instructions(functions)

        end_tag = "\n<|eot_id|>"
        return llama_message + system_message + end_tag

    def add_tool_instructions(self, tools: list) -> str:
        tool_prompt = """
    You have access to the following tools:
    """

        for tool in tools:
            if tool["type"] == "function":
                func = tool["function"]
                tool_prompt += (
                    f"Use the function '{func['name']}' to '{func['description']}':\n"
                )
                params_info = json.dumps(func["parameters"], indent=4)
                tool_prompt += f"Parameters format:\n{params_info}\n\n"

        tool_prompt += """
If you choose to use a function to produce this response, ONLY reply in the following format with no prefix or suffix:
§{"type": "function", "name": "FUNCTION_NAME", "parameters": {"PARAMETER_NAME": PARAMETER_VALUE}}
IMPORTANT: IT IS VITAL THAT YOU NEVER ADD A PREFIX OR A SUFFIX TO THE FUNCTION CALL.

Here is an example of the output I desiere when performing function call:
§{"type": "function", "name": "python_repl_ast", "parameters": {"query": "print(df.shape)"}}
NOTE: There is no prefix before the symbol '§' and nothing comes after the call is done.

    Reminder:
    - Function calls MUST follow the specified format.
    - Only call one function at a time.
    - Required parameters MUST be specified.
    - Put the entire function call reply on one line.
    - If there is no function call available, answer the question like normal with your current knowledge and do not tell the user about function calls.
    - If you have already called a tool and got the response for the users question please reply with the response.
    """

        return tool_prompt

    def add_function_instructions(self, functions: list) -> str:
        function_prompt = """
You have access to the following functions:
"""

        for func in functions:
            function_prompt += (
                f"Use the function '{func['name']}' to: '{func['description']}'\n"
            )
            params_info = json.dumps(func["parameters"], indent=4)
            function_prompt += f"{params_info}\n\n"

        function_prompt += """
If you choose to use a function to produce this response, ONLY reply in the following format with no prefix or suffix:
§{"type": "function", "name": "FUNCTION_NAME", "parameters": {"PARAMETER_NAME": PARAMETER_VALUE}}

Here is an example of the output I desiere when performing function call:
§{"type": "function", "name": "python_repl_ast", "parameters": {"query": "print(df.shape)"}}

Reminder:
- Function calls MUST follow the specified format.
- Only call one function at a time.
- NEVER call more than one function at a time.
- Required parameters MUST be specified.
- Put the entire function call reply on one line.
- If there is no function call available, answer the question like normal with your current knowledge and do not tell the user about function calls.
- If you have already called a function and got the response for the user's question, please reply with the response.
"""
        return function_prompt

    def add_conversation(self, openai_message: list, llama_message: str) -> str:
        conversation_parts = []
        for message in openai_message:
            if message["role"] == "system":
                continue
            elif message["role"] == "user" and isinstance(message["content"], str):
                try:
                    # Attempt to safely evaluate the string to a Python object
                    content_as_list = ast.literal_eval(message["content"])
                    if isinstance(content_as_list, list):
                        # If the content is a list, process each nested message
                        for nested_message in content_as_list:
                            conversation_parts.append(
                                self.format_message(nested_message)
                            )
                    else:
                        # If the content is not a list, append it directly
                        conversation_parts.append(self.format_message(message))
                except (ValueError, SyntaxError):
                    # If evaluation fails or content is not a list/dict string, append the message directly
                    conversation_parts.append(self.format_message(message))
            else:
                # For all other messages, use the existing formatting logic
                conversation_parts.append(self.format_message(message))

        return llama_message + "".join(conversation_parts)

    def format_message(self, message: dict) -> str:
        """Format a single message for the conversation."""
        if "tool_calls" in message:
            for tool_call in message["tool_calls"]:
                function_name = tool_call["function"]["name"]
                arguments = tool_call["function"]["arguments"]
                return f"""
        <|start_header_id|>assistant<|end_header_id|>
        <function={function_name}>{arguments}</function>
        <|eom_id|>
        """
        elif "tool_call_id" in message:
            tool_response = message["content"]
            return f"""
    <|start_header_id|>ipython<|end_header_id|>
    {tool_response}
    <|eot_id|>
    """
        elif "function_call" in message:
            function_name = message["function_call"]["name"]
            arguments = message["function_call"]["arguments"]
            return f"""
        <|start_header_id|>assistant<|end_header_id|>
        <function={function_name}>{arguments}</function>
        <|eom_id|>
        """
        elif (
            message["role"] in ["assistant", "user"] and message["content"] is not None
        ):
            return f"""
    <|start_header_id|>{message['role']}<|end_header_id|>
    {message['content']}
    <|eot_id|>
    """
        elif message["role"] == "function":
            function_response = message["content"]
            return f"""
    <|start_header_id|>ipython<|end_header_id|>
    {function_response}
    <|eot_id|>
    """
        return ""



import time
import uuid
from abc import ABC, abstractmethod
from typing import (
    Any,
    AsyncGenerator,
    Coroutine,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    Union,
)

import tiktoken
from llmstudio_core.exceptions import ProviderError
from openai.types.chat import (
    ChatCompletion,
    ChatCompletionChunk,
    ChatCompletionMessage,
    ChatCompletionMessageToolCall,
)
from openai.types.chat.chat_completion import Choice
from openai.types.chat.chat_completion_message import FunctionCall
from openai.types.chat.chat_completion_message_tool_call import Function
from pydantic import BaseModel, ValidationError

provider_registry = {}


def provider(cls):
    """Decorator to register a new provider."""
    provider_registry[cls._provider_config_name()] = cls

    return cls


class ChatRequest(BaseModel):
    chat_input: Any
    model: str
    is_stream: Optional[bool] = False
    retries: Optional[int] = 0
    parameters: Optional[dict] = {}

    def __init__(self, **data):
        super().__init__(**data)
        base_model_fields = self.model_fields.keys()
        additional_params = {
            k: v for k, v in data.items() if k not in base_model_fields
        }
        self.parameters.update(additional_params)


class Provider(ABC):
    END_TOKEN = "<END_TOKEN>"

    def __init__(
        self,
        config: Any,
        api_key: Optional[str] = None,
        api_endpoint: Optional[str] = None,
        api_version: Optional[str] = None,
        base_url: Optional[str] = None,
        tokenizer: Optional[Any] = None,
    ):
        self.config = config
        self.API_KEY = api_key
        self.api_endpoint = api_endpoint
        self.api_version = api_version
        self.base_url = base_url

        self.tokenizer = tokenizer if tokenizer else self._get_tokenizer()
        self.count = 0

    @abstractmethod
    async def achat(
        self,
        chat_input: Any,
        model: str,
        is_stream: Optional[bool] = False,
        retries: Optional[int] = 0,
        parameters: Optional[dict] = {},
        **kwargs,
    ) -> Coroutine[Any, Any, Union[ChatCompletionChunk, ChatCompletion]]:
        raise NotImplementedError("Providers needs to have achat method implemented.")

    @abstractmethod
    def chat(
        self,
        chat_input: Any,
        model: str,
        is_stream: Optional[bool] = False,
        retries: Optional[int] = 0,
        parameters: Optional[dict] = {},
        **kwargs,
    ) -> Union[ChatCompletionChunk, ChatCompletion]:
        raise NotImplementedError("Providers needs to have chat method implemented.")

    @staticmethod
    @abstractmethod
    def _provider_config_name():
        raise NotImplementedError(
            "Providers need to implement the '_provider_config_name' property."
        )


class ProviderCore(Provider):
    END_TOKEN = "<END_TOKEN>"

    @abstractmethod
    def validate_request(self, request: ChatRequest):
        raise NotImplementedError("Providers need to implement the 'validate_request'.")

    @abstractmethod
    async def agenerate_client(
        self, request: ChatRequest
    ) -> Coroutine[Any, Any, Generator]:
        """Generate the provider's client"""
        raise NotImplementedError("Providers need to implement the 'agenerate_client'.")

    @abstractmethod
    def generate_client(self, request: ChatRequest) -> Generator:
        """Generate the provider's client"""
        raise NotImplementedError("Providers need to implement the 'generate_client'.")

    @abstractmethod
    async def aparse_response(self, response: AsyncGenerator, **kwargs) -> Any:
        raise NotImplementedError("ProviderCore needs a aparse_response method.")

    @abstractmethod
    def parse_response(self, response: AsyncGenerator, **kwargs) -> Any:
        raise NotImplementedError("ProviderCore needs a parse_response method.")

    def validate_model(self, request: ChatRequest):
        if request.model not in self.config.models:
            raise ProviderError(
                f"Model {request.model} is not supported by {self.config.name}"
            )

    async def achat(
        self,
        chat_input: Any,
        model: str,
        is_stream: Optional[bool] = False,
        retries: Optional[int] = 0,
        parameters: Optional[dict] = {},
        **kwargs,
    ):

        """Makes a chat connection with the provider's API"""
        try:
            request = self.validate_request(
                dict(
                    chat_input=chat_input,
                    model=model,
                    is_stream=is_stream,
                    retries=retries,
                    parameters=parameters,
                    **kwargs,
                )
            )
        except ValidationError as e:
            raise ProviderError(str(e))

        self.validate_model(request)

        for _ in range(request.retries + 1):
            try:
                start_time = time.time()
                response = await self.agenerate_client(request)
                response_handler = self.ahandle_response(request, response, start_time)

                if request.is_stream:
                    return response_handler
                else:
                    return await response_handler.__anext__()
            # except HTTPException as e:
            #     if e.status_code == 429:
            #         continue  # Retry on rate limit error
            #     else:
            #         raise e  # Raise other HTTP exceptions
            except Exception as e:
                raise ProviderError(str(e))
        raise ProviderError("Too many requests")

    def chat(
        self,
        chat_input: Any,
        model: str,
        is_stream: Optional[bool] = False,
        retries: Optional[int] = 0,
        parameters: Optional[dict] = {},
        **kwargs,
    ):

        """Makes a chat connection with the provider's API"""
        try:
            request = self.validate_request(
                dict(
                    chat_input=chat_input,
                    model=model,
                    is_stream=is_stream,
                    retries=retries,
                    parameters=parameters,
                    **kwargs,
                )
            )
        except ValidationError as e:
            raise ProviderError(str(e))

        self.validate_model(request)

        for _ in range(request.retries + 1):
            try:
                start_time = time.time()
                response = self.generate_client(request)
                response_handler = self.handle_response(request, response, start_time)

                if request.is_stream:
                    return response_handler
                else:
                    return response_handler.__next__()
            # except HTTPExceptio as e:
            #     if e.status_code == 429:
            #         continue  # Retry on rate limit error
            #     else:
            #         raise e  # Raise other HTTP exceptions
            except Exception as e:
                raise ProviderError(str(e))
        raise ProviderError("Too many requests")

    async def ahandle_response(
        self, request: ChatRequest, response: AsyncGenerator, start_time: float
    ) -> AsyncGenerator[str, None]:
        """Handles the response from an API"""
        first_token_time = None
        previous_token_time = None
        token_times = []
        token_count = 0
        chunks = []

        async for chunk in self.aparse_response(response, request=request):
            token_count += 1
            current_time = time.time()
            first_token_time = first_token_time or current_time
            if previous_token_time is not None:
                token_times.append(current_time - previous_token_time)
            previous_token_time = current_time

            chunks.append(chunk)
            if request.is_stream:
                chunk = chunk[0] if isinstance(chunk, tuple) else chunk
                model = chunk.get("model")
                if chunk.get("choices")[0].get("finish_reason") != "stop":
                    chat_output = chunk.get("choices")[0].get("delta").get("content")
                    chunk = {
                        **chunk,
                        "id": str(uuid.uuid4()),
                        "chat_input": (
                            request.chat_input
                            if isinstance(request.chat_input, str)
                            else request.chat_input[-1]["content"]
                        ),
                        "chat_output": None,
                        "chat_output_stream": chat_output if chat_output else "",
                        "context": (
                            [{"role": "user", "content": request.chat_input}]
                            if isinstance(request.chat_input, str)
                            else request.chat_input
                        ),
                        "provider": self.config.id,
                        "model": (
                            request.model
                            if model and model.startswith(request.model)
                            else (model or request.model)
                        ),
                        "deployment": (
                            model
                            if model and model.startswith(request.model)
                            else (request.model if model != request.model else None)
                        ),
                        "timestamp": time.time(),
                        "parameters": request.parameters,
                        "metrics": None,
                    }
                    yield ChatCompletionChunk(**chunk)

        chunks = [chunk[0] if isinstance(chunk, tuple) else chunk for chunk in chunks]
        model = next(chunk["model"] for chunk in chunks if chunk.get("model"))

        response, output_string = self.join_chunks(chunks, request)

        metrics = self.calculate_metrics(
            request.chat_input,
            response,
            request.model,
            start_time,
            time.time(),
            first_token_time,
            token_times,
            token_count,
        )

        response = {
            **(chunk if request.is_stream else response.model_dump()),
            "id": str(uuid.uuid4()),
            "chat_input": (
                request.chat_input
                if isinstance(request.chat_input, str)
                else request.chat_input[-1]["content"]
            ),
            "chat_output": output_string,
            "chat_output_stream": "",
            "context": (
                [{"role": "user", "content": request.chat_input}]
                if isinstance(request.chat_input, str)
                else request.chat_input
            ),
            "provider": self.config.id,
            "model": (
                request.model
                if model and model.startswith(request.model)
                else (model or request.model)
            ),
            "deployment": (
                model
                if model and model.startswith(request.model)
                else (request.model if model != request.model else None)
            ),
            "timestamp": time.time(),
            "parameters": request.parameters,
            "metrics": metrics,
        }

        if request.is_stream:
            yield ChatCompletionChunk(**response)
        else:
            yield ChatCompletion(**response)

    def handle_response(
        self, request: ChatRequest, response: Generator, start_time: float
    ) -> Generator:
        """Handles the response from an API"""
        first_token_time = None
        previous_token_time = None
        token_times = []
        token_count = 0
        chunks = []

        for chunk in self.parse_response(response, request=request):
            token_count += 1
            current_time = time.time()
            first_token_time = first_token_time or current_time
            if previous_token_time is not None:
                token_times.append(current_time - previous_token_time)
            previous_token_time = current_time

            chunks.append(chunk)
            if request.is_stream:
                chunk = chunk[0] if isinstance(chunk, tuple) else chunk
                model = chunk.get("model")
                if chunk.get("choices")[0].get("finish_reason") != "stop":
                    chat_output = chunk.get("choices")[0].get("delta").get("content")
                    chunk = {
                        **chunk,
                        "id": str(uuid.uuid4()),
                        "chat_input": (
                            request.chat_input
                            if isinstance(request.chat_input, str)
                            else request.chat_input[-1]["content"]
                        ),
                        "chat_output": None,
                        "chat_output_stream": chat_output if chat_output else "",
                        "context": (
                            [{"role": "user", "content": request.chat_input}]
                            if isinstance(request.chat_input, str)
                            else request.chat_input
                        ),
                        "provider": self.config.id,
                        "model": (
                            request.model
                            if model and model.startswith(request.model)
                            else (model or request.model)
                        ),
                        "deployment": (
                            model
                            if model and model.startswith(request.model)
                            else (request.model if model != request.model else None)
                        ),
                        "timestamp": time.time(),
                        "parameters": request.parameters,
                        "metrics": None,
                    }
                    yield ChatCompletionChunk(**chunk)

        chunks = [chunk[0] if isinstance(chunk, tuple) else chunk for chunk in chunks]
        model = next(chunk["model"] for chunk in chunks if chunk.get("model"))

        response, output_string = self.join_chunks(chunks, request)

        metrics = self.calculate_metrics(
            request.chat_input,
            response,
            request.model,
            start_time,
            time.time(),
            first_token_time,
            token_times,
            token_count,
        )

        response = {
            **(chunk if request.is_stream else response.model_dump()),
            "id": str(uuid.uuid4()),
            "chat_input": (
                request.chat_input
                if isinstance(request.chat_input, str)
                else request.chat_input[-1]["content"]
            ),
            "chat_output": output_string,
            "chat_output_stream": "",
            "context": (
                [{"role": "user", "content": request.chat_input}]
                if isinstance(request.chat_input, str)
                else request.chat_input
            ),
            "provider": self.config.id,
            "model": (
                request.model
                if model and model.startswith(request.model)
                else (model or request.model)
            ),
            "deployment": (
                model
                if model and model.startswith(request.model)
                else (request.model if model != request.model else None)
            ),
            "timestamp": time.time(),
            "parameters": request.parameters,
            "metrics": metrics,
        }

        if request.is_stream:
            yield ChatCompletionChunk(**response)
        else:
            yield ChatCompletion(**response)

    def join_chunks(self, chunks, request):

        finish_reason = chunks[-1].get("choices")[0].get("finish_reason")
        if finish_reason == "tool_calls":
            tool_calls = {}
            for chunk in chunks:
                try:
                    data = chunk.get("choices")[0].get("delta").get("tool_calls")[0]
                    tool_calls.setdefault(data["index"], []).append(data)
                except TypeError:
                    continue

            tool_call_ids = [t[0].get("id") for t in tool_calls.values()]
            tool_call_names = [
                t[0].get("function").get("name") for t in tool_calls.values()
            ]
            tool_call_types = [
                t[0].get("function").get("type", "function")
                for t in tool_calls.values()
            ]

            tool_call_arguments_all = []
            for t in tool_calls.values():
                tool_call_arguments_all.append(
                    "".join(
                        chunk.get("function", {}).get("arguments", "") for chunk in t
                    )
                )

            tool_calls_parsed = [
                ChatCompletionMessageToolCall(
                    id=tool_call_id,
                    function=Function(
                        arguments=tool_call_arguments, name=tool_call_name
                    ),
                    type=tool_call_type,
                )
                for tool_call_arguments, tool_call_name, tool_call_type, tool_call_id in zip(
                    tool_call_arguments_all,
                    tool_call_names,
                    tool_call_types,
                    tool_call_ids,
                )
            ]

            try:
                return (
                    ChatCompletion(
                        id=chunks[-1].get("id"),
                        created=chunks[-1].get("created"),
                        model=chunks[-1].get("model"),
                        object="chat.completion",
                        choices=[
                            Choice(
                                finish_reason="tool_calls",
                                index=0,
                                logprobs=None,
                                message=ChatCompletionMessage(
                                    content=None,
                                    role="assistant",
                                    function_call=None,
                                    tool_calls=tool_calls_parsed,
                                ),
                            )
                        ],
                    ),
                    str(tool_call_names + tool_call_arguments_all),
                )
            except Exception as e:
                raise e
        elif finish_reason == "function_call":
            function_calls = [
                chunk.get("choices")[0].get("delta").get("function_call")
                for chunk in chunks[:-1]
                if chunk.get("choices")
                and chunk.get("choices")[0].get("delta")
                and chunk.get("choices")[0].get("delta").get("function_call")
            ]

            function_call_name = function_calls[0].get("name")

            function_call_arguments = ""
            for chunk in function_calls:
                function_call_arguments += chunk.get("arguments")

            return (
                ChatCompletion(
                    id=chunks[-1].get("id"),
                    created=chunks[-1].get("created"),
                    model=chunks[-1].get("model"),
                    object="chat.completion",
                    choices=[
                        Choice(
                            finish_reason="function_call",
                            index=0,
                            logprobs=None,
                            message=ChatCompletionMessage(
                                content=None,
                                role="assistant",
                                tool_calls=None,
                                function_call=FunctionCall(
                                    arguments=function_call_arguments,
                                    name=function_call_name,
                                ),
                            ),
                        )
                    ],
                ),
                function_call_arguments,
            )

        elif finish_reason == "stop" or finish_reason == "length":
            if self.__class__.__name__ in ("OpenAIProvider", "AzureProvider"):
                start_index = 1
            else:
                start_index = 0

            stop_content = "".join(
                filter(
                    None,
                    [
                        chunk.get("choices")[0].get("delta").get("content")
                        for chunk in chunks[start_index:]
                    ],
                )
            )

            return (
                ChatCompletion(
                    id=chunks[-1].get("id"),
                    created=chunks[-1].get("created"),
                    model=chunks[-1].get("model"),
                    object="chat.completion",
                    choices=[
                        Choice(
                            finish_reason="stop",
                            index=0,
                            logprobs=None,
                            message=ChatCompletionMessage(
                                content=stop_content,
                                role="assistant",
                                function_call=None,
                                tool_calls=None,
                            ),
                        )
                    ],
                ),
                stop_content,
            )

    def calculate_metrics(
        self,
        input: Any,
        output: Any,
        model: str,
        start_time: float,
        end_time: float,
        first_token_time: float,
        token_times: Tuple[float, ...],
        token_count: int,
    ) -> Dict[str, Any]:
        """Calculates metrics based on token times and output"""
        model_config = self.config.models[model]
        input_tokens = len(self.tokenizer.encode(self.input_to_string(input)))
        output_tokens = len(self.tokenizer.encode(self.output_to_string(output)))

        input_cost = self.calculate_cost(input_tokens, model_config.input_token_cost)
        output_cost = self.calculate_cost(output_tokens, model_config.output_token_cost)

        total_time = end_time - start_time
        return {
            "input_tokens": input_tokens,
            "output_tokens": output_tokens,
            "total_tokens": input_tokens + output_tokens,
            "cost_usd": input_cost + output_cost,
            "latency_s": total_time,
            "time_to_first_token_s": first_token_time - start_time,
            "inter_token_latency_s": sum(token_times) / len(token_times),
            "tokens_per_second": token_count / total_time,
        }

    def calculate_cost(
        self, token_count: int, token_cost: Union[float, List[Dict[str, Any]]]
    ) -> float:
        if isinstance(token_cost, list):
            for cost_range in token_cost:
                if token_count >= cost_range.range[0] and (
                    token_count <= cost_range.range[1] or cost_range.range[1] is None
                ):
                    return cost_range.cost * token_count
        else:
            return token_cost * token_count
        return 0

    def input_to_string(self, input):
        if isinstance(input, str):
            return input
        else:
            result = []
            for message in input:
                if message.get("content") is not None:
                    if isinstance(message["content"], str):
                        result.append(message["content"])
                    elif (
                        isinstance(message["content"], list)
                        and message.get("role") == "user"
                    ):
                        for item in message["content"]:
                            if item.get("type") == "text":
                                result.append(item.get("text", ""))
                            elif item.get("type") == "image_url":
                                url = item.get("image_url", {}).get("url", "")
                                result.append(url)
            return "".join(result)

    def output_to_string(self, output):
        if output.choices[0].finish_reason == "stop":
            return output.choices[0].message.content
        elif output.choices[0].finish_reason == "tool_calls":
            return output.choices[0].message.tool_calls[0].function.arguments
        elif output.choices[0].finish_reason == "function_call":
            return output.choices[0].message.function_call.arguments

    def get_end_token_string(self, metrics: Dict[str, Any]) -> str:
        return f"{self.END_TOKEN},input_tokens={metrics['input_tokens']},output_tokens={metrics['output_tokens']},cost_usd={metrics['cost_usd']},latency_s={metrics['latency_s']:.5f},time_to_first_token_s={metrics['time_to_first_token_s']:.5f},inter_token_latency_s={metrics['inter_token_latency_s']:.5f},tokens_per_second={metrics['tokens_per_second']:.2f}"

    def _get_tokenizer(self):
        return {}.get(self.config.id, tiktoken.get_encoding("cl100k_base"))



# from llmstudio_core.providers.anthropic import AnthropicProvider #TODO: adpat it
from typing import Optional

from llmstudio_core.providers.azure import AzureProvider

# from llmstudio_core.providers.ollama import OllamaProvider #TODO: adapt it
from llmstudio_core.providers.openai import OpenAIProvider
from llmstudio_core.providers.provider import ProviderCore, provider_registry
from llmstudio_core.providers.vertexai import VertexAIProvider
from llmstudio_core.utils import _load_providers_config

_engine_config = _load_providers_config()


def LLMCore(provider: str, api_key: Optional[str] = None, **kwargs) -> ProviderCore:
    """
    Factory method to create an instance of a provider.

    Args:
        provider (str): The name of the provider.
        api_key (Optional[str], optional): The API key for the provider. Defaults to None.

    Returns:
        ProviderCore: An instance of the provider.

    Raises:
        NotImplementedError: If the provider is not found in the provider map.
    """
    provider_config = _engine_config.providers.get(provider)
    provider_class = provider_registry.get(provider_config.id)
    if provider_class:
        return provider_class(config=provider_config, api_key=api_key, **kwargs)
    raise NotImplementedError(
        f"Provider not found: {provider_config.id}. Available providers: {str(provider_registry.keys())}"
    )
